﻿''' <summary>
''' Accepts DragObjects and displays their images at the location the user drop them.
''' The user can select the images, and then nudge, rotate, or drag them elsewhere.
''' </summary>
Public Class DropTarget
    Inherits Control

    ' most subs here are overrides.

#Region " Variables "
    ' DroppedImages is our collection of droppedImage - it stores the final placement
    ' for all the images dropped into the workspace
    Public droppedImages As List(Of DroppedImage)
    Private m_dragObject As DragObject
    Public indexOfLastImageClicked As Integer
    Private dragging As Boolean
    Private m_SelectionPen As Pen
#End Region

#Region " Constructor "
    Public Sub New()
        indexOfLastImageClicked = -1
        Me.SetStyle(ControlStyles.AllPaintingInWmPaint Or ControlStyles.OptimizedDoubleBuffer _
                    Or ControlStyles.UserPaint, True)
        Me.UpdateStyles()
        Me.AllowDrop = True
        droppedImages = New List(Of DroppedImage)
        m_SelectionPen = New Pen(Color.Black)
        m_SelectionPen.DashStyle = Drawing2D.DashStyle.Dash
    End Sub
#End Region

#Region " Dispose "
    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        If m_dragObject IsNot Nothing Then m_dragObject.Dispose()
        For Each droppedItem As DroppedImage In droppedImages
            If droppedItem IsNot Nothing Then droppedItem.Dispose()
        Next
        If m_SelectionPen IsNot Nothing Then m_SelectionPen.Dispose()
        MyBase.Dispose(disposing)
    End Sub
#End Region

#Region " Protected Overrides "

#Region " IsInputKey "
    Protected Overrides Function IsInputKey(ByVal keyData As System.Windows.Forms.Keys) As Boolean
        ' some keys don't raise the keydown event by default
        Select Case keyData
            Case Keys.Left, Keys.Right, Keys.Up, Keys.Down, Keys.Control
                Return True
            Case Else
                Return MyBase.IsInputKey(keyData)
        End Select
    End Function

#End Region

#Region " OnDragDrop "

    Protected Overrides Sub OnDragDrop(ByVal drgevent As System.Windows.Forms.DragEventArgs)
        If dragging Then dragging = False
        ' we no longer need to give feedback.
        RemoveHandler m_dragObject.GiveFeedback, AddressOf DropTargetControl_GiveFeedback
        ' Calculate the point to draw the image.
        ' The hotspot of the gdi cursor is in the center of the image.
        Dim dropPoint As Point = Me.PointToClient(New Point(drgevent.X, drgevent.Y))
        dropPoint.X = dropPoint.X - (m_dragObject.Image.Size.Width / 2)
        dropPoint.Y = dropPoint.Y - (m_dragObject.Image.Size.Height / 2)
        ' create a new DroppedImage instance
        Dim droppedImage1 As New DroppedImage(m_dragObject.Image.Clone, dropPoint)
        ' Deselect any selected image
        ClearSelection()
        droppedImage1.Selected = True
        ' Add it to the end of our list.  It will be drawn last, and so appear on top.
        droppedImages.Add(droppedImage1)
        ' Invalidate the control so the new image is drawn.
        Me.Invalidate(SlightlyBiggerRectangle(droppedImage1.BoundingRectangle))
        Application.DoEvents()
        m_dragObject.Dispose()
        Me.indexOfLastImageClicked = droppedImages.Count - 1
        ' the drag initiator still has focus
        Me.Focus()
        MyBase.OnDragDrop(drgevent)
    End Sub

#End Region

#Region " OnDragEnter "

    Protected Overrides Sub OnDragEnter(ByVal drgevent As System.Windows.Forms.DragEventArgs)
        ' is the data being dragged in one of our dragObject types?
        If drgevent.Data.GetDataPresent(GetType(DragObject)) Then
            ' Get the dragObject, chain the feedback events together
            m_dragObject = drgevent.Data.GetData(GetType(DragObject))
            AddHandler m_dragObject.GiveFeedback, AddressOf DropTargetControl_GiveFeedback
            drgevent.Effect = DragDropEffects.Copy
        Else
            m_dragObject = Nothing
            drgevent.Effect = DragDropEffects.None
        End If

        MyBase.OnDragEnter(drgevent)
    End Sub

#End Region

#Region " OnDragLeave "

    Protected Overrides Sub OnDragLeave(ByVal e As System.EventArgs)
        RemoveHandler m_dragObject.GiveFeedback, AddressOf DropTargetControl_GiveFeedback
        MyBase.OnDragLeave(e)
    End Sub

#End Region

#Region " OnKeyDown "

    Protected Overrides Sub OnKeyDown(ByVal e As System.Windows.Forms.KeyEventArgs)
        ' we want to be able to rotate (?) the image, nudge the image by (one?) pixel
        ' no diagonals (I don't have any on my laptop), and expand and shrink the image.  

        If dragging Then Exit Sub
        If indexOfLastImageClicked = -1 Then Exit Sub

        Select Case e.KeyData
            ' The keys enumeration is a <flags> enumeration, so when the left key and the control key
            ' are pressed TOGETHER the value is the biwise OR of the two key values:
            Case Keys.Left Or Keys.Control
                ' rotate counter-clockwise
                With droppedImages(indexOfLastImageClicked)
                    ' Tell the droppedImage to go rotate itself
                    .RotateCounterClockwise()
                    ' then invalidate
                    Me.Invalidate(SlightlyBiggerRectangle(.BoundingRectangle))
                End With
            Case Keys.Left
                ' nudge left.
                ' TODO: full selections
                With droppedImages(indexOfLastImageClicked)
                    ' don't let the user drag the thing miles out of bounds...
                    If .Location.X + .Image.Width < 1 Then Exit Select
                    ' nudge it left:
                    .Location = New Point(.Location.X - 1, .Location.Y)
                    ' now invalidate
                    Me.Invalidate(SlightlyBiggerRectangle(.BoundingRectangle))
                    ' call application.doevent, so that the invalidate fires immediately
                End With
            Case Keys.Right Or Keys.Control
                ' rotate clockwise
                With droppedImages(indexOfLastImageClicked)
                    .RotateClockwise()
                    Me.Invalidate(SlightlyBiggerRectangle(.BoundingRectangle))
                End With
            Case Keys.Right
                ' nudge right.
                With droppedImages(indexOfLastImageClicked)
                    If .Location.X > Me.Width - 1 Then Exit Select
                    .Location = New Point(.Location.X + 1, .Location.Y)
                    Me.Invalidate(SlightlyBiggerRectangle(.BoundingRectangle))
                End With
            Case Keys.Up
                ' nudge up.
                ' TODO:  We might not really need up and down.  Use up to make the image size bigger?
                With droppedImages(indexOfLastImageClicked)
                    If .Location.Y + +.Image.Width < 1 Then Exit Select
                    .Location = New Point(.Location.X, .Location.Y - 1)
                    Me.Invalidate(SlightlyBiggerRectangle(.BoundingRectangle))
                End With
            Case Keys.Down
                ' nudge down.
                ' TODO:  Maybe make this shrink the image size?
                With droppedImages(indexOfLastImageClicked)
                    If .Location.Y > Me.Height - 1 Then Exit Select
                    .Location = New Point(.Location.X, .Location.Y + 1)
                    Me.Invalidate(SlightlyBiggerRectangle(.BoundingRectangle))
                End With
            Case Keys.Delete
                Dim rec As Rectangle = droppedImages(indexOfLastImageClicked).BoundingRectangle
                droppedImages.RemoveAt(indexOfLastImageClicked)
                indexOfLastImageClicked = -1
                Me.Invalidate(SlightlyBiggerRectangle(rec))
        End Select

        MyBase.OnKeyDown(e)
    End Sub

#End Region

#Region " OnLostFocus "

    Protected Overrides Sub OnLostFocus(ByVal e As System.EventArgs)
        If indexOfLastImageClicked > -1 Then
            Me.Refresh()
            indexOfLastImageClicked = -1
            dragging = False
        End If
        MyBase.OnLostFocus(e)
    End Sub

#End Region

#Region " OnMouseDown "

    Protected Overrides Sub OnMouseDown(ByVal e As System.Windows.Forms.MouseEventArgs)
        If droppedImages.Count = 0 Then Exit Sub
        ' if an image was previously selected, then de-select it
        ClearSelection()
        ' search backwards, check each droppedImage to see if the point is inside the image...
        Dim indexOfClickedImage As Integer = -1
        For index As Integer = droppedImages.Count - 1 To 0 Step -1
            ' lazy test  - use a 1x1 rectangle and check for intersection
            If droppedImages(index).BoundingRectangle.IntersectsWith(New Rectangle(e.X, e.Y, 1, 1)) Then
                indexOfClickedImage = index
                Exit For
            End If
        Next
        If indexOfClickedImage > -1 Then
            indexOfLastImageClicked = indexOfClickedImage
            ' if the user is selecting, then we do the select on mouse up
            ' if the user is dragging, then we initiate the drag once the mouse has moved
            ' The fact that m_indexOfLastImageClicked is not -1 indicates that something is happening.
            ' We also need to determine the point (in image co-ordinates) clicked.
            With droppedImages(indexOfLastImageClicked)
                Dim x As Integer = e.Location.X - .Location.X
                Dim y As Integer = e.Location.Y - .Location.Y
                .PointClicked = New Point(x, y)
            End With
        End If
        MyBase.OnMouseDown(e)
    End Sub

#End Region

#Region " OnMouseMove "

    Protected Overrides Sub OnMouseMove(ByVal e As System.Windows.Forms.MouseEventArgs)
        ' if an item is selected, and the button is down, adn the drag hasn't already started...
        If Not dragging AndAlso e.Button = Windows.Forms.MouseButtons.Left AndAlso indexOfLastImageClicked > -1 Then

            ' the user wants to start a drag drop
            Dim imageBeingDragged As DragObject
            Dim rec As Rectangle
            With droppedImages(indexOfLastImageClicked)
                imageBeingDragged = New DragObject(.RotatedImage.Clone)
                rec = .BoundingRectangle
                .Dispose()
            End With
            ' remove this item for the list
            droppedImages.RemoveAt(indexOfLastImageClicked)
            ' refresh the image to remove it...
            Me.Invalidate(SlightlyBiggerRectangle(rec))
            ' reset the indexOfLastImageClicked
            indexOfLastImageClicked = -1
            ' start dragdrop
            AddHandler Me.GiveFeedback, AddressOf imageBeingDragged.RaiseGiveFeedback
            Me.DoDragDrop(imageBeingDragged, DragDropEffects.Copy)
        End If
        MyBase.OnMouseMove(e)
    End Sub

#End Region

#Region " OnMouseUp "

    Protected Overrides Sub OnMouseUp(ByVal e As System.Windows.Forms.MouseEventArgs)
        ' uninterested unless the user has just clicked to select an image
        If dragging OrElse indexOfLastImageClicked = -1 Then Exit Sub
        ' the user wants to select an image.
        ' set all droppedImages to unselected.
        ClearSelection()
        ' set the one that was clicked to selected.
        droppedImages(indexOfLastImageClicked).Selected = True
        ' redraw
        Me.Refresh()
        MyBase.OnMouseUp(e)
    End Sub

#End Region

#Region " OnPaint "

    Protected Overrides Sub OnPaint(ByVal e As System.Windows.Forms.PaintEventArgs)
        MyBase.OnPaint(e)
        Dim g As Graphics = e.Graphics
        ' draw all the droppedImages...
        For Each itemToDraw As DroppedImage In droppedImages
            e.Graphics.DrawImage(itemToDraw.RotatedImage, itemToDraw.BoundingRectangle)
            If itemToDraw.Selected Then
                g.DrawRectangle(m_SelectionPen, itemToDraw.BoundingRectangle)
            End If
        Next

    End Sub

#End Region

#End Region

#Region " Private "

#Region " Methods "

    ''' <summary>
    ''' Use the image as cursor when drag-dropping.
    ''' </summary>
    Private Sub DropTargetControl_GiveFeedback(ByVal e As GiveFeedbackEventArgs)
        If e.Effect = DragDropEffects.Copy Then
            e.UseDefaultCursors = False
            Windows.Forms.Cursor.Current = m_dragObject.ImageCursor
        Else
            e.UseDefaultCursors = True
            Windows.Forms.Cursor.Current = Cursors.No
        End If
    End Sub

    ''' <summary>
    ''' Returns a slightly bigger rectangle than the one passed in.
    ''' Note:  When you invalidate a rectangle it always misses a bit.
    ''' </summary> 
    Public Function SlightlyBiggerRectangle(ByVal recIn As Rectangle) As Rectangle
        Return New Rectangle(recIn.X - 2, recIn.Y - 2, recIn.Width + 5, recIn.Height + 5)
    End Function

#End Region

#End Region

#Region " Public "

#Region " Methods "
    ''' <summary>
    ''' clear the selection rectangle.
    ''' </summary> 
    Public Sub ClearSelection()
        For Each di As DroppedImage In droppedImages
            If di.Selected = True Then
                di.Selected = False
                Me.Invalidate(SlightlyBiggerRectangle(di.BoundingRectangle))
            End If
        Next
    End Sub

#End Region

#End Region

End Class
